import copy
import itertools
import numpy as np
import scipy
import SVAR.SVARutilGMM
from scipy.optimize import LinearConstraint


def innovation(u, b, restrictions=[], whiten=False, blocks=False ):
    # Calculates unmixed innovations according to: e = B^{-1} u

    B = get_BMatrix(b, restrictions=restrictions, whiten=whiten, blocks=blocks )
    A = np.linalg.inv(B)
    e = np.matmul(A, np.transpose(u))
    e = np.transpose(e)
    return e

def get_BMatrix(b, restrictions=[], whiten=False, blocks=False):
    # Transforms vectorized B into B
    if whiten:
        # ToDo: Add white restrictions
        B = get_Orthogonal(b, blocks)
    else:
        if blocks:
            restrictions = getRestrictions_blocks(blocks)

        if np.array(restrictions).size == 0:
            b_length = np.sqrt(np.size(b))
            n = b_length.astype(int)
            restrictions = np.full([n, n], np.nan)

        B = copy.deepcopy(restrictions)
        B[np.isnan(B)] = b

    return B

def get_BVector(B, restrictions=[], whiten=False, blocks=False):
    # inverse of get_BMatrix
    # ToDo: Error if element of B!=restrictions
    if whiten:
        # ToDo: Add restrictions
        b = get_Skewsym(B, blocks=blocks)
    else:
        if np.array(restrictions).size == 0:
            restrictions = np.full(np.shape(B), np.nan)
        b = B[np.isnan(restrictions) == 1]
    return b

def get_block_rec(n_rec, n):
    if not (n_rec):
        n_rec = 0
    blocks = list()
    for i in range(1, n_rec + 1):
        blocks.append(np.array([i, i]))
    if n_rec + 1 <= n:
        blocks.append(np.array([n_rec + 1, n]))
    return blocks

def getRestrictions_recursive(B):
    # restricts upper right triangular matrix
    myrestrictions = np.full(np.shape(B), np.nan)
    iu1 = np.triu_indices(np.shape(myrestrictions)[1], 1)
    myrestrictions[iu1] = B[iu1]
    return myrestrictions


def getRestrictions_blocks(blocks):
    myrestrictions = np.full([blocks[-1][-1], blocks[-1][-1]], np.nan)
    for block in blocks:
        myrestrictions[block[0] - 1:block[1], block[1]:] = 0
    return myrestrictions

def getRestrictions_none(n):
    myrestrictions = np.full([n,n], np.nan)
    return myrestrictions

def do_whitening(u, white):
    if white:
        T, n = np.shape(u)
        Sigma = np.dot(np.transpose(u), u) / T
        V = np.linalg.cholesky(Sigma)
    else:
        V = np.eye(np.size(u))
    Vinv = np.linalg.inv(V)
    z = np.matmul(Vinv, np.transpose(u))
    z = np.transpose(z)
    return z, V


def get_Skewsym(B, blocks=False):
    if not blocks:
        n = int(np.shape(B)[0])
        blocks = list()
        blocks.append(np.array([1, n]))
    s = np.array([])
    for block in blocks:
        n = block[1] - block[0] + 1
        B_this = B[block[0] - 1:block[1], block[0] - 1:block[1]]
        S = scipy.linalg.logm(B_this)
        il1 = np.tril_indices(n, k=-1)
        s = np.append(s,S[il1])

    return s


def get_Orthogonal(b, blocks=False):
    if not blocks:
        n = int(np.ceil(np.sqrt(2 * np.size(b))))
        blocks = list()
        blocks.append(np.array([1, n]))
    B = np.eye(blocks[-1][-1])

    b_counter = 0
    for block in blocks:
        n = block[1] - block[0] + 1
        S = np.full([n, n], 0.0)
        il1 = np.tril_indices(n, k=-1)
        n_s = int(np.size(il1) / 2)
        S[il1] = b[b_counter:b_counter + n_s]
        S = S - np.transpose(S)
        # B_this = scipy.linalg.expm(S)
        B_this, expm_frechet_AE = scipy.linalg.expm_frechet(S, S)
        B[block[0] - 1:block[1], block[0] - 1:block[1]] = B_this
        b_counter = b_counter + n_s
    if not (b_counter == np.shape(b)[0]):
        raise ValueError('Specified b value does not match B restrictions.')
    return B


## Permutation
def PermToB0(B0,B,avar,restrictions,T):
    b = get_BVector(B, restrictions=restrictions, whiten=False)
    n = np.shape(B0)[0]
    R = np.eye(np.shape(b)[0])
    avar = avar[np.logical_not(np.isnan(avar))]
    avar = np.reshape(avar, [np.size(b), np.size(b)])
    perms = get_AllSignPerm(n)
    score = np.zeros(np.shape(perms)[0])
    for i, perm in enumerate(perms):
        B0perm = np.matmul(B0, perm)
        if (restrictions[np.isnan(restrictions) == False] == B0perm[np.isnan(restrictions) == False]).all():
            r =get_BVector(B0perm, restrictions=restrictions, whiten=False)
            waldstat, wald_p = SVAR.SVARutilGMM.wald(R, r, b, avar, T)
            score[i] = wald_p
    Bbest = np.matmul(B, np.linalg.inv(perms[np.argmax(score)]))
    return Bbest, perms[np.argmax(score)]



def get_AllSignPerm(n):
    Perms = []

    permutations = list(itertools.permutations(range(n)))
    num_permutations = np.shape(permutations)[0]

    signs = list(itertools.product(np.array([1, -1]), repeat=n))
    num_signs = np.shape(signs)[0]

    I = np.eye(n)
    for p in range(num_permutations):
        perm_this = I[permutations[p], :]
        for s in range(num_signs):
            sign_this = np.matmul(I, np.diag(signs[s]))
            sign_perm_this = np.matmul(perm_this, sign_this)
            Perms.append(sign_perm_this)

    return Perms


def get_Omega(e):
    n = np.shape(e)[1]
    CoOmega2 = np.full([n, n], np.nan)
    CoOmega3 = np.full([n, n, n], np.nan)
    CoOmega4 = np.full([n, n, n, n], np.nan)
    for pi in range(n):
        for pj in range(pi, n):
            save_ij = np.array([np.prod(e[:, np.array([pi, pj])], axis=1)]).T
            CoOmega2[pi, pj] = np.mean(save_ij)
            for pk in range(pj, n):
                save_ijk = np.array([np.prod(np.append(save_ij, e[:, np.array([pk])], axis=1), axis=1)]).T
                CoOmega3[pi, pj, pk] = np.mean(save_ijk)
                for pl in range(pk, n):
                    save_ijkl = np.prod(np.append(save_ijk, e[:, np.array([pl])], axis=1), axis=1)
                    CoOmega4[pi, pj, pk, pl] = np.mean(save_ijkl)
    return CoOmega2, CoOmega3, CoOmega4

def get_CoOmega(e ):
    [T,n] = np.shape(e)
    CoOmega1 = np.mean(e.T,axis=1)
    CoOmega2 = np.matmul(e.T , e   )/T
    CoOmega3 = np.full([n, n, n], np.nan)
    CoOmega4 = np.full([n, n, n, n], np.nan)
    CoOmega5 = np.full([n, n, n, n,n], np.nan)
    CoOmega6 = np.full([n, n, n, n,n,n], np.nan)
    CoOmega7 = np.full([n, n, n, n,n,n,n], np.nan)
    CoOmega8 = np.full([n, n, n, n,n,n,n,n], np.nan)
    for p3 in range(n):
        e_save3 = e.T * e[:,p3]
        CoOmega3[:,:,p3] = np.matmul( e_save3    , e    )/T
        for p4 in range(n):
            e_save4 = e_save3  * e[:,p4]
            CoOmega4[:,:,p3,p4] = np.matmul(e_save4 , e  )/T
            for p5 in range(n):
                e_save5 = e_save4 * e[:,p5]
                CoOmega5[:,:,p3,p4,p5] = np.matmul(e_save5 , e  )/T
                for p6 in range(n):
                    e_save6 = e_save5 * e[:,p6]
                    CoOmega6[:, :, p3, p4, p5, p6] = np.matmul(e_save6, e ) / T
                    for p7 in range(n):
                        e_save7 = e_save6 * e[:,p7]
                        CoOmega7[:, :, p3, p4, p5, p6, p7] = np.matmul(e_save7, e ) / T
                        for p8 in range(n):
                            e_save8 = e_save7 * e[:,p8]
                            CoOmega8[:, :, p3, p4, p5, p6, p7, p8] = np.matmul(e_save8, e ) / T

    return CoOmega1, CoOmega2, CoOmega3, CoOmega4, CoOmega5, CoOmega6, CoOmega7, CoOmega8

def get_Omega_Moments(e):
    n = np.shape(e)[1]
    Omega = np.full([n, 6], np.nan)
    Omega[:, 0] = np.mean(np.power(e, 1), axis=0)
    Omega[:, 1] = np.mean(np.power(e, 2), axis=0)
    Omega[:, 2] = np.mean(np.power(e, 3), axis=0)
    Omega[:, 3] = np.mean(np.power(e, 4), axis=0)
    Omega[:, 4] = np.mean(np.power(e, 5), axis=0)
    Omega[:, 5] = np.mean(np.power(e, 6), axis=0)

    return Omega


def get_omegaext(e):
    n = np.shape(e)[1]
    omegaext = np.full([n, 7,n+1], np.nan)

    omegaext[:, 0,0] = np.mean(np.power(e, 0), axis=0)
    omegaext[:, 1,0] = np.mean(np.power(e, 1), axis=0)
    omegaext[:, 2,0] = np.mean(np.power(e, 2), axis=0)
    omegaext[:, 3,0] = np.mean(np.power(e, 3), axis=0)
    omegaext[:, 4,0] = np.mean(np.power(e, 4), axis=0)
    omegaext[:, 5,0] = np.mean(np.power(e, 5), axis=0)
    omegaext[:, 6,0] = np.mean(np.power(e, 6), axis=0)

    for i in range(n):
        omegaext[:, 0, i+1] = np.mean(np.multiply(np.power(e, 0), e[:,i].reshape(-1, 1)), axis=0)
        omegaext[:, 1, i+1] = np.mean(np.multiply(np.power(e, 1), e[:,i].reshape(-1, 1) ), axis=0)
        omegaext[:, 2, i+1] = np.mean(np.multiply(np.power(e, 2), e[:,i].reshape(-1, 1) ), axis=0)
        omegaext[:, 3, i+1] = np.mean(np.multiply(np.power(e, 3), e[:,i].reshape(-1, 1) ), axis=0)
        omegaext[:, 4, i+1] = np.mean(np.multiply(np.power(e, 4), e[:,i].reshape(-1, 1) ), axis=0)
        omegaext[:, 5, i+1] = np.mean(np.multiply(np.power(e, 5), e[:,i].reshape(-1, 1) ), axis=0)
        omegaext[:, 6, i+1] = np.mean(np.multiply(np.power(e, 6), e[:,i].reshape(-1, 1) ), axis=0)

    return omegaext
 


import numpy as np
import SVAR
from SVAR.SVARutilGMM import get_Moments,  get_S_uncentered, get_g_wf,get_G_Indep, get_S_Indep,get_S
import SVAR.estimatorGMMW
import scipy.stats

def loss(u, b, restrictions, moments, moments_powerindex, blocks=False):
    whiten = True
    g = get_g_wf(b=b, u=u, restrictions=restrictions, moments=moments, moments_powerindex=moments_powerindex,
              whiten=whiten, blocks=blocks)
    Q = - np.sum(np.power(g, 2))
    return Q

def loss_MIcorrection(u,b,restrictions,moments, moments_powerindex, blocks=False):
    whiten = True
    g = get_g_wf(b=b, u=u, restrictions=restrictions, moments=moments, moments_powerindex=moments_powerindex,
              whiten=whiten, blocks=blocks )
    Q = + 6* np.sum(np.power(g, 2))

    return Q

def gradient(u, b, Jacobian, W, restrictions, moments, moments_powerindex):
    # e = SVAR.innovation(u, b, restrictions=restrictions)
    # CoOmega = SVAR.SVARutil.get_CoOmega(e)
    #
    # Jac_temp = Jacobian(u=u, b=b, restrictions=restrictions, CoOmega=CoOmega)
    # this_g = get_g(b=b[:], u=u[:], restrictions=restrictions[:],
    #                moments=moments[:], moments_powerindex=moments_powerindex)
    # dGMM_temp  =  2 * np.linalg.multi_dot([np.transpose(Jac_temp), W, this_g])
    return []

def get_AVAR(n, G2,GC, S22, S2C, SC2, SCC, Wfast, restrictions=[] , blocks=[] ):
    # ToDo: only works with default moments
    V = SVAR.estimatorGMMW.get_Avar(n, G2,GC, S22, S2C, SC2, SCC, W=Wfast,restrictions=restrictions)

    return V


def prepareOptions(u,
                   addThirdMoments=True, addFourthMoments=True, moments=[], moments_blocks=True, moments_MeanIndep=False,
                   bstart=[], bstartopt='Rec',
                   Avarparametric='Uncorrelated',
                   blocks=False, n_rec=False,
                   printOutput=True,
                   estimator = 'GMM_WF',
                   Bcetenter=[]
                   ):
    options = dict()


    options['estimator'] = estimator

    T, n = np.shape(u)
    options['T'] = T
    options['n'] = n

    options['printOutput'] = printOutput

    options['whiten'] = True
    _, V = SVAR.do_whitening(u, white=True)
    options['V'] = V


    if Avarparametric == 'Uncorrelated' or Avarparametric == 'Uncorrelated_uncentered' or Avarparametric == 'Independent':
        options['Avarparametric'] = Avarparametric
    else:
        print("Invalid Avarparametric value. Set Avarparametric=Uncorrelated")
        options['Avarparametric'] = 'Uncorrelated'

    restrictions, blocks = SVAR.estPrepare.prepare_blocks_restrictions(n, n_rec, blocks, restrictions=[])
    options['restrictions'] = restrictions
    options['blocks'] = blocks

    options['addThirdMoments'] = addThirdMoments
    options['addFourthMoments'] = addFourthMoments
    moments = SVAR.estPrepare.prepare_moments(estimator, moments, addThirdMoments, addFourthMoments, moments_blocks,
                                              blocks, n)
    options['moments'] = moments
    options['moments_powerindex'] = SVAR.SVARutilGMM.get_Moments_powerindex(moments)

    options['moments_MeanIndep'] = moments_MeanIndep
    if moments_MeanIndep:
        options['moments_MIcorrection'] = SVAR.SVARutilGMM.get_Moments_MIcorrection(n, blocks=blocks)
        options['moments_MIcorrection_powerindex'] = SVAR.SVARutilGMM.get_Moments_powerindex(options['moments_MIcorrection'])

    bstart = SVAR.estPrepare.prepare_bstart(estimator, bstart, u, options, bstartopt=bstartopt)
    options['bstart'] = bstart

    options['Bcetenter'] = Bcetenter


    return options


def SVARout(est_SVAR, options, u):
    T, n = np.shape(u)
    out_SVAR = dict()
    out_SVAR['options'] = options

    b_est = est_SVAR['x']
    out_SVAR['b_est'] = b_est

    B_est = SVAR.get_BMatrix(b_est, restrictions=options['restrictions'], whiten=options['whiten'],
                             blocks=options['blocks'])
    B_est = np.matmul(options['V'], B_est)
    out_SVAR['B_est'] = B_est

    e = SVAR.innovation(u, SVAR.get_BVector(B_est, restrictions=options['restrictions'],blocks=options['blocks']), restrictions=options['restrictions'] ,
                        blocks=options['blocks'])
    out_SVAR['e'] = e

    Omega_all = SVAR.SVARutil.get_Omega(e)
    out_SVAR['Omega_all'] = Omega_all
    omega = SVAR.SVARutil.get_Omega_Moments(e)
    CoOmega = SVAR.SVARutil.get_CoOmega(e)
    out_SVAR['omega'] = omega

    z, options['V'] = SVAR.do_whitening(u, white=True)
    out_SVAR['loss'] = loss(z, b_est, options['restrictions'], options['moments'], options['moments_powerindex'], blocks=options['blocks'])
    if options['moments_MeanIndep']:
        out_SVAR['loss'] = out_SVAR['loss'] - SVAR.estimatorGMMWF.loss_MIcorrection(z,b_est,restrictions=options['restrictions'],
                                                                  moments=options['moments_MIcorrection'],
                                                                  moments_powerindex=options['moments_MIcorrection_powerindex'], blocks=options['blocks'])

    Moments2 = SVAR.SVARutilGMM.get_Mr(2, n)
    Moments2 = np.append(Moments2, SVAR.SVARutilGMM.get_Cr(2, n), axis=0)
    comoments = get_Moments('GMM_W', n, blocks= options['blocks'],
                            addThirdMoments= options['addThirdMoments'],
                            addFourthMoments= options['addFourthMoments'],moments_MeanIndep=options['moments_MeanIndep'])
    Wfast = SVAR.estimatorGMMW.get_W_fast(comoments)
    if options['Avarparametric'] == 'Independent':
        G2 = get_G_Indep(Moments2, out_SVAR['B_est'], out_SVAR['omega'], options['restrictions'])
        GC = get_G_Indep(comoments, out_SVAR['B_est'], out_SVAR['omega'], options['restrictions'])

        S22 = get_S_Indep(Moments2, Moments2, out_SVAR['omega'])
        S2C = get_S_Indep(Moments2, comoments, out_SVAR['omega'])
        SC2 = get_S_Indep(comoments, Moments2, out_SVAR['omega'])
        SCC = get_S_Indep(comoments, comoments, out_SVAR['omega'])
    elif options['Avarparametric'] == 'Uncorrelated_uncentered':
        Jacobian = SVAR.SVARutilGMM.generate_Jacobian_function(moments=comoments, restrictions=options['restrictions'])
        Jacobian2 = SVAR.SVARutilGMM.generate_Jacobian_function(moments=Moments2, restrictions=options['restrictions'])

        GC = Jacobian(u=u, b= SVAR.get_BVector(B_est,restrictions=options['restrictions']) , restrictions=options['restrictions'], CoOmega=CoOmega)
        G2 = Jacobian2(u=u, b= SVAR.get_BVector(B_est,restrictions=options['restrictions']) , restrictions=options['restrictions'], CoOmega=CoOmega)

        if np.shape(comoments)[0] == 0:
            GC = get_G_Indep(comoments, out_SVAR['B_est'], out_SVAR['omega'], options['restrictions'])

        S = get_S_uncentered(u, SVAR.get_BVector(B_est, restrictions=np.full([n, n], np.nan)),
                             np.append(Moments2, comoments, axis=0), np.full([n, n], np.nan))
        S22 = S[:np.shape(Moments2)[0],:np.shape(Moments2)[0]]
        S2C = S[:np.shape(Moments2)[0],np.shape(Moments2)[0]:]
        SC2 = S[np.shape(Moments2)[0]:,:np.shape(Moments2)[0]]
        SCC = S[np.shape(Moments2)[0]:,np.shape(Moments2)[0]:]
    elif options['Avarparametric'] == 'Uncorrelated':
        Jacobian = SVAR.SVARutilGMM.generate_Jacobian_function(moments=comoments, restrictions=options['restrictions'])
        Jacobian2 = SVAR.SVARutilGMM.generate_Jacobian_function(moments=Moments2, restrictions=options['restrictions'])

        GC = Jacobian(u=u, b= SVAR.get_BVector(B_est,restrictions=options['restrictions']) , restrictions=options['restrictions'], CoOmega=CoOmega)
        G2 = Jacobian2(u=u, b= SVAR.get_BVector(B_est,restrictions=options['restrictions']) , restrictions=options['restrictions'], CoOmega=CoOmega)

        if np.shape(comoments)[0] == 0:
            GC = get_G_Indep(comoments, out_SVAR['B_est'], out_SVAR['omega'], options['restrictions'])

        S = get_S(u, SVAR.get_BVector(B_est,restrictions=np.full([n, n], np.nan)),  np.append(Moments2, comoments, axis=0) , np.full([n, n], np.nan) )
        S22 = S[:np.shape(Moments2)[0],:np.shape(Moments2)[0]]
        S2C = S[:np.shape(Moments2)[0],np.shape(Moments2)[0]:]
        SC2 = S[np.shape(Moments2)[0]:,:np.shape(Moments2)[0]]
        SCC = S[np.shape(Moments2)[0]:,np.shape(Moments2)[0]:]

    V_est = get_AVAR(n, G2,GC, S22, S2C, SC2, SCC ,restrictions=options['restrictions'] , Wfast=Wfast)
    out_SVAR['Avar_est'] = V_est


    b = SVAR.get_BVector(B_est, restrictions=options['restrictions'], whiten=False)
    t_all = np.empty_like(b)
    t_all_p = np.empty_like(b)
    for idx, bthis in np.ndenumerate(b):
        avar = out_SVAR['Avar_est']
        avar = avar[np.logical_not(np.isnan(avar))]
        avar = np.reshape(avar, [np.size(b), np.size(b)])
        avar_this = avar[idx, idx]

        t_this = T * np.divide(np.power(b[idx], 2), avar_this)
        t_all[idx] = t_this
        t_all_p[idx] = 1 - scipy.stats.chi2.cdf(t_all[idx], 1)
        # cdf_this = scipy.stats.norm.cdf(t_this)
        # if cdf_this < 0.5:
        #     t_all_p[idx] = 2 * (scipy.stats.norm.cdf(t_this))
        # else:
        #     t_all_p[idx] = 2*(1-scipy.stats.norm.cdf(t_this))
    out_SVAR['t_all'] = t_all
    out_SVAR['t_all_p'] = t_all_p


    # Parameter wald
    out_SVAR['wald_all'], out_SVAR['wald_all_p'] = SVAR.SVARutilGMM.wald_param_all(B_est, options['restrictions'], out_SVAR['Avar_est'], T)

    # Fval - Rec
    out_SVAR['wald_rec'], out_SVAR['wald_rec_p'] = SVAR.SVARutilGMM.waldRec(B_est, out_SVAR['Avar_est'],
                                                                            options['restrictions'], T)


    if options['printOutput']:
            SVAR.estOutput.print_out(n, T, out_SVAR)

    return out_SVAR



import setuptools
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")
license_text = (this_directory / "LICENSE").read_text(encoding="utf-8")

# Read requirements.txt, handling FileNotFoundError
try:
  requirements = (this_directory / "requirements.txt").read_text(encoding="utf-8").splitlines()
except FileNotFoundError:
  requirements = []
  print("Warning: requirements.txt not found. Proceeding without dependencies.")

setuptools.setup(
  name="SVARpy",
  version="0.1.10",
  author="Sascha Keweloh",
  author_email="sascha.keweloh@tu-dortmund.de",
  description="SVAR estimation",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/Saschakew/SVARpy",
  packages=setuptools.find_packages(),
  install_requires=requirements,
  python_requires='>=3.7',
  license=license_text,
  classifiers=[
      "Programming Language :: Python :: 3",
      "License :: OSI Approved :: MIT License",
      "Operating System :: OS Independent",
  ],
)



import numpy as np
import SVAR
from SVAR.SVARutilGMM import get_Moments,  get_S_uncentered,  get_g, get_G_Indep, get_S_Indep, get_S






def loss(u, b, W, restrictions, moments, moments_powerindex, blocks=False):
    whiten = True
    g = get_g(b=b, u=u, restrictions=restrictions, moments=moments, moments_powerindex=moments_powerindex,
              whiten=whiten, blocks=blocks)
    Q = np.linalg.multi_dot([g, W, g])
    return Q

def gradient(u, b, Jacobian, W, restrictions, moments, moments_powerindex):
    e = SVAR.innovation(u, b, restrictions=restrictions,whiten=True)
    CoOmega = SVAR.SVARutil.get_CoOmega(e)

    Jac_temp = Jacobian(u=u, b=b, restrictions=restrictions, CoOmega=CoOmega)

    this_g = get_g(b=b[:], u=u[:], restrictions=restrictions[:],
                   moments=moments[:], moments_powerindex=moments_powerindex)

    dGMM_temp  =  2 * np.linalg.multi_dot([np.transpose(Jac_temp), W, this_g])
    return dGMM_temp

def get_W_fast(Moments):
    W = np.eye(np.shape(Moments)[0])

    counter = 0
    for moment in Moments:
        r = np.sum(moment)

        tmp = np.zeros(np.shape(moment))
        counter_in = 0
        for i in moment:
            tmp[counter_in] = np.math.factorial(moment[counter_in])
            counter_in += 1

        W[counter, counter] = np.math.factorial(r) / np.prod(tmp)
        counter += 1

    return W

def get_Avar(n, G2,GC, S22, S2C, SC2, SCC, W=[],restrictions=[]):
    Moments2 = SVAR.SVARutilGMM.get_Mr(2, n)
    Moments2 = np.append(Moments2, SVAR.SVARutilGMM.get_Cr(2, n), axis=0)

    if np.array(W).size == 0:
        W = np.linalg.inv(SCC)

    if np.array(restrictions).size == 0:
        restrictions = np.full([n, n], np.nan)
    number_of_restrictions = np.sum(restrictions == 0)


    H11 = G2
    H12 = np.zeros([np.shape(Moments2)[0], np.shape(Moments2)[0]])
    H21 = np.matmul(np.matmul(np.transpose(GC), W), GC)
    H22 = - np.transpose(G2)
    H = np.hstack((np.vstack((H11, H21)), np.vstack((H12, H22))))
    H = np.linalg.inv(H)
    M2 = H[0:(n * n - number_of_restrictions), 0: np.shape(Moments2)[0]]
    MC_tmp1 = H[0: (n * n - number_of_restrictions), np.shape(Moments2)[0]:]
    MC_tmp2 = np.matmul(np.transpose(GC), W)
    MC = np.matmul(MC_tmp1, MC_tmp2)
    V1 = np.matmul(np.matmul(M2, S22), np.transpose(M2))
    V2 = np.matmul(np.matmul(M2, S2C), np.transpose(MC))
    V3 = np.matmul(np.matmul(MC, SC2), np.transpose(M2))
    V4 = np.matmul(np.matmul(MC, SCC), np.transpose(MC))
    V_est = V1 + V2 + V3 + V4

    elementcounter = 0
    for i in range(n):
        for j in range(n):
            if not (np.isnan(restrictions[i, j])):
                V_est = np.insert(V_est, elementcounter, np.full(np.shape(V_est)[1], np.NaN), 0)
                V_est = np.insert(V_est, elementcounter, np.full(np.shape(V_est)[0], np.NaN), 1)
            elementcounter += 1
    return V_est

def get_GMM_W_Avar_param(Moments, B, omega, restrictions=[], W=[]):
    # Calculates analytically the asymptotic variance of sqrt(T)(b_hat-b) with b_hat GMM estimator with optimal weighting

    n, n = np.shape(B)

    if np.array(restrictions).size == 0: restrictions = np.full([n, n], np.nan)
    number_of_restrictions = np.sum(restrictions == 0)

    Moments2 = SVAR.SVARutilGMM.get_Mr(2, n)
    Moments2 = np.append(Moments2, SVAR.SVARutilGMM.get_Cr(2, n), axis=0)

    G2 = get_G_Indep(Moments2, B, omega, restrictions)
    GC = get_G_Indep(Moments, B, omega, restrictions)

    S22 = get_S_Indep(Moments2, Moments2, omega)
    S2C = get_S_Indep(Moments2, Moments, omega)
    SC2 = get_S_Indep(Moments, Moments2, omega)
    SCC = get_S_Indep(Moments, Moments, omega)

    if np.array(W).size == 0:  W = np.linalg.inv(SCC)

    H11 = G2
    H12 = np.zeros([np.shape(Moments2)[0], np.shape(Moments2)[0]])
    H21 = np.matmul(np.matmul(np.transpose(GC), W), GC)
    H22 = - np.transpose(G2)
    H = np.hstack((np.vstack((H11, H21)), np.vstack((H12, H22))))
    H = np.linalg.inv(H)
    M2 = H[0:(n * n - number_of_restrictions), 0: np.shape(Moments2)[0]]
    MC_tmp1 = H[0: (n * n - number_of_restrictions), np.shape(Moments2)[0]:]
    MC_tmp2 = np.matmul(np.transpose(GC), W)
    MC = np.matmul(MC_tmp1, MC_tmp2)
    V1 = np.matmul(np.matmul(M2, S22), np.transpose(M2))
    V2 = np.matmul(np.matmul(M2, S2C), np.transpose(MC))
    V3 = np.matmul(np.matmul(MC, SC2), np.transpose(M2))
    V4 = np.matmul(np.matmul(MC, SCC), np.transpose(MC))
    V = V1 + V2 + V3 + V4

    # np.linalg.det(np.matmul(np.transpose(G2),G2))
    # tmp=np.matmul(GC,np.transpose(GC))
    # test = np.linalg.inv(np.matmul(np.matmul(np.transpose(GC), tmp), GC))
    if False:
        H11 = np.matmul(np.matmul(np.transpose(GC), W), GC)
        H12 = - np.transpose(G2)
        H21 = G2
        H22 = np.zeros([np.shape(Moments2)[0], np.shape(Moments2)[0]])
        H = np.hstack((np.vstack((H11, H21)), np.vstack((H12, H22))))
        K = np.linalg.inv(H)
        K11 = K[:(n * n - number_of_restrictions),:(n * n - number_of_restrictions)]
        K12 = K[:(n * n - number_of_restrictions), (n * n - number_of_restrictions): ]
        K21 = K[(n * n - number_of_restrictions): ,:(n * n - number_of_restrictions)]
        K22 = K[(n * n - number_of_restrictions): , (n * n - number_of_restrictions):  ]


        tmp = np.linalg.inv(np.matmul(    np.transpose(G2) ,G2))
        tmp2 = - np.matmul( G2,tmp) # =K21
        - np.matmul(K21,np.matmul(    np.transpose(G2) ,G2)) #=G2


        tmp = np.linalg.inv(np.matmul(G2, np.transpose(G2)))
        tmp2 = - np.matmul( np.matmul( np.matmul(np.matmul(np.transpose(GC), W), GC), np.transpose(G2)), tmp)
        tmp3 = np.matmul(K21,tmp2 ) # = K22



        tmp =  np.matmul(np.matmul(np.matmul(np.transpose(GC), W), GC), np.transpose( np.matmul(np.matmul(np.transpose(GC), W), GC) ))
        np.linalg.det(tmp)



    # invGWG = np.linalg.inv(np.matmul(np.matmul(np.transpose(GC), W), GC))
    # invG2invGWGG2 = np.linalg.inv(np.matmul(np.matmul(G2, invGWG), np.transpose(G2)))
    # L1 = np.matmul(np.matmul(invGWG, np.transpose(G2)), invG2invGWGG2)
    # L2 = - np.matmul(L1, np.matmul(np.matmul(G2, invGWG), np.matmul(np.transpose(GC), W)))
    # L3 = np.matmul(invGWG, np.matmul(np.transpose(GC), W))
    # L4 = L2+L3
    # R1 = np.matmul(np.matmul(L1, S22), np.transpose(L1))
    # R2 = np.matmul(np.matmul(L1, S2C), np.transpose(L4))
    # R3 = np.matmul(np.matmul(L4, SC2), np.transpose(L1))
    # R4 = np.matmul(np.matmul(L4, SCC), np.transpose(L4))
    # R = R1 + R2 + R3 + R4


    elementcounter = 0
    for i in range(n):
        for j in range(n):
            if not (np.isnan(restrictions[i, j])):
                V = np.insert(V, elementcounter, np.full(np.shape(V)[1], np.NaN), 0)
                V = np.insert(V, elementcounter, np.full(np.shape(V)[0], np.NaN), 1)
            elementcounter += 1

    return V



def prepareOptions(u,
                   addThirdMoments=True, addFourthMoments=True, moments=[], moments_blocks=True,   moments_MeanIndep=False,
                   bstart=[], bstartopt='Rec',
                   Avarparametric='Uncorrelated',
                   blocks=False, n_rec=False,
                   W=[], Wstartopt='I',
                   printOutput=True,
                    estimator = 'GMM_W'
                   ):
    options = dict()


    options['estimator'] = estimator

    T, n = np.shape(u)
    options['T'] = T
    options['n'] = n

    options['printOutput'] = printOutput

    options['whiten'] = True
    _, V = SVAR.do_whitening(u, white=True)
    options['V'] = V

    if Avarparametric == 'Uncorrelated' or Avarparametric == 'Independent' or Avarparametric == 'Uncorrelated_uncentered':
        options['Avarparametric'] = Avarparametric
    else:
        print("Invalid Avarparametric value. Set Avarparametric=Uncorrelated")
        options['Avarparametric'] = 'Uncorrelated'

    restrictions, blocks = SVAR.estPrepare.prepare_blocks_restrictions(n, n_rec, blocks, restrictions=[])
    options['restrictions'] = restrictions
    options['blocks'] = blocks

    moments = SVAR.estPrepare.prepare_moments('GMM_W', moments, addThirdMoments, addFourthMoments, moments_blocks,
                                              blocks, n, moments_MeanIndep)
    options['moments'] = moments
    options['moments_MeanIndep'] = moments_MeanIndep
    options['moments_powerindex'] = SVAR.SVARutilGMM.get_Moments_powerindex(moments)

    bstart = SVAR.estPrepare.prepare_bstart(estimator, bstart, u, options, bstartopt=bstartopt)
    options['bstart'] = bstart


    Jacobian = SVAR.SVARutilGMM.generate_Jacobian_function(moments=moments, restrictions=restrictions)
    options['Jacobian'] = Jacobian

    # Weighting
    if Wstartopt == 'I':
        W = np.eye(np.shape(moments)[0])
    elif Wstartopt == 'GMM_WF':
        W = get_W_fast(moments)
    elif Wstartopt == 'specific':
        W = W
    else:
        raise ValueError('Unknown option Wstartopt')
    options['W'] = W

    if np.shape(options['moments'])[0] < np.shape(options['bstart'])[0]:
        raise ValueError('Less moment conditions than parameters. The SVAR is not identified')

    return options


def SVARout(est_SVAR, options, u):
    T, n = np.shape(u)
    out_SVAR = dict()
    out_SVAR['options'] = options

    b_est = est_SVAR['x']
    out_SVAR['b_est'] = b_est

    B_est = SVAR.get_BMatrix(b_est, restrictions=options['restrictions'], whiten=options['whiten'],
                             blocks=options['blocks'])
    B_est = np.matmul(options['V'], B_est)
    out_SVAR['B_est'] = B_est

    e = SVAR.innovation(u, SVAR.get_BVector(B_est, restrictions=options['restrictions']),
                        restrictions=options['restrictions'],
                        blocks=options['blocks'])
    out_SVAR['e'] = e

    Omega_all = SVAR.SVARutil.get_Omega(e)
    out_SVAR['Omega_all'] = Omega_all
    omega = SVAR.SVARutil.get_Omega_Moments(e)
    CoOmega = SVAR.SVARutil.get_CoOmega(e)
    out_SVAR['omega'] = omega

    # out_SVAR['loss'] = est_SVAR['fun']
    z, options['V'] = SVAR.do_whitening(u, white=True)
    out_SVAR['loss'] = loss(z, b_est, options['W'], options['restrictions'], options['moments'], options['moments_powerindex'], blocks=options['blocks'])





    Moments2 = SVAR.SVARutilGMM.get_Mr(2, n)
    Moments2 = np.append(Moments2, SVAR.SVARutilGMM.get_Cr(2, n), axis=0)
    if options['Avarparametric'] == 'Independent':
        G2 = get_G_Indep(Moments2, out_SVAR['B_est'], out_SVAR['omega'], options['restrictions'])
        GC = get_G_Indep(options['moments'], out_SVAR['B_est'], out_SVAR['omega'], options['restrictions'])

        S22 = get_S_Indep(Moments2, Moments2, out_SVAR['omega'])
        S2C = get_S_Indep(Moments2, options['moments'], out_SVAR['omega'])
        SC2 = get_S_Indep(options['moments'], Moments2, out_SVAR['omega'])
        SCC = get_S_Indep(options['moments'], options['moments'], out_SVAR['omega'])
    elif options['Avarparametric'] == 'Uncorrelated':
        Jacobian = SVAR.SVARutilGMM.generate_Jacobian_function(moments=options['moments'], restrictions=options['restrictions'])
        Jacobian2 = SVAR.SVARutilGMM.generate_Jacobian_function(moments=Moments2, restrictions=options['restrictions'])

        GC = Jacobian(u=u, b= SVAR.get_BVector(B_est,restrictions=options['restrictions']) , restrictions=options['restrictions'], CoOmega=CoOmega)
        G2 = Jacobian2(u=u, b= SVAR.get_BVector(B_est,restrictions=options['restrictions']) , restrictions=options['restrictions'], CoOmega=CoOmega)

        if np.shape(options['moments'])[0] == 0:
            GC = get_G_Indep(options['moments'], out_SVAR['B_est'], out_SVAR['omega'], options['restrictions'])

        S = get_S(u, SVAR.get_BVector(B_est,restrictions=np.full([n, n], np.nan)),  np.append(Moments2, options['moments'], axis=0) , np.full([n, n], np.nan) )
        S22 = S[:np.shape(Moments2)[0],:np.shape(Moments2)[0]]
        S2C = S[:np.shape(Moments2)[0],np.shape(Moments2)[0]:]
        SC2 = S[np.shape(Moments2)[0]:,:np.shape(Moments2)[0]]
        SCC = S[np.shape(Moments2)[0]:,np.shape(Moments2)[0]:]
    elif options['Avarparametric'] == 'Uncorrelated_uncentered':
        Jacobian = SVAR.SVARutilGMM.generate_Jacobian_function(moments=options['moments'],
                                                               restrictions=options['restrictions'])
        Jacobian2 = SVAR.SVARutilGMM.generate_Jacobian_function(moments=Moments2, restrictions=options['restrictions'])

        GC = Jacobian(u=u, b=SVAR.get_BVector(B_est, restrictions=options['restrictions']),
                      restrictions=options['restrictions'], CoOmega=CoOmega)
        G2 = Jacobian2(u=u, b=SVAR.get_BVector(B_est, restrictions=options['restrictions']),
                       restrictions=options['restrictions'], CoOmega=CoOmega)

        if np.shape(options['moments'])[0] == 0:
            GC = get_G_Indep(options['moments'], out_SVAR['B_est'], out_SVAR['omega'], options['restrictions'])
        S = get_S_uncentered(u, SVAR.get_BVector(B_est,restrictions=np.full([n, n], np.nan)),  np.append(Moments2, options['moments'], axis=0) , np.full([n, n], np.nan) )
        S22 = S[:np.shape(Moments2)[0],:np.shape(Moments2)[0]]
        S2C = S[:np.shape(Moments2)[0],np.shape(Moments2)[0]:]
        SC2 = S[np.shape(Moments2)[0]:,:np.shape(Moments2)[0]]
        SCC = S[np.shape(Moments2)[0]:,np.shape(Moments2)[0]:]

    V_est = get_Avar(n, G2, GC, S22, S2C, SC2, SCC, W=options['W'], restrictions=options['restrictions'])
    out_SVAR['Avar_est'] = V_est



    # V_est = get_GMM_W_Avar_param(options['moments'], B=out_SVAR['B_est'], omega=out_SVAR['omega'],
    #                                restrictions=options['restrictions'], W=options['W'])



    # Parameter wald
    out_SVAR['wald_all'], out_SVAR['wald_all_p'] = SVAR.SVARutilGMM.wald_param_all(B_est, options['restrictions'], out_SVAR['Avar_est'], T)

    # Fval - Rec
    out_SVAR['wald_rec'], out_SVAR['wald_rec_p'] = SVAR.SVARutilGMM.waldRec(B_est, out_SVAR['Avar_est'],
                                                                            options['restrictions'], T)


    if options['printOutput']:
            SVAR.estOutput.print_out(n, T, out_SVAR)


    return out_SVAR


import numpy as np
import pandas as pd
from tabulate import tabulate
import SVAR

def genOutput(estimator, u, Bbest, options):
    if options['whiten']:
        Bbest = np.matmul(np.linalg.inv(options['V']), Bbest)
    est_SVAR = dict()
    est_SVAR['x'] = SVAR.get_BVector(Bbest, options['restrictions'],
                                     whiten=options['whiten'],
                                     blocks=options['blocks'])
    if estimator == 'GMM' or estimator == 'CUE' or estimator == 'CSUE':
        GMM_out = SVAR.estimatorGMM.SVARout(est_SVAR, options, u)
    elif estimator == 'GMM_W':
        GMM_out = SVAR.estimatorGMMW.SVARout(est_SVAR, options, u)
    elif estimator == 'GMM_WF':
        GMM_out = SVAR.estimatorGMMWF.SVARout(est_SVAR, options, u)
    elif estimator == 'PML':
        GMM_out = SVAR.estimatorPML.SVARout(est_SVAR, options, u)
    else:
        print("Unknown estimator in genOutput")

    return GMM_out

def idx_Best(n):
    idx = np.array([])
    for i in np.arange(n):
        string = "B(" + str(i + 1) + ',:)'
        idx = np.append(idx, string)
    return idx


def print_Avar(V):
    n = int(np.sqrt(np.shape(V)[0]))
    print(pd.DataFrame(data=np.reshape(np.diag(V), [n, n]),
                       index=idx_Best(n),
                       columns=col_Best(  n)))


def print_Moments(omega):
    n = np.shape(omega)[0]
    omega = np.round(omega,2)
    omega = omega[:,1:-2]
    print(pd.DataFrame(data=omega, index=idx_moments(n),
                       columns=np.array([  'E[e^2]', 'E[e^3]', 'E[e^4]' ])))


def print_B(B):
    n = np.shape(B)[0]
    print(pd.DataFrame(data=B, index=idx_Best(n),
                       columns=col_Best(n)))


def idx_moments(n):
    idx = np.array([])
    for i in np.arange(n):
        string = "e" + str(i + 1)
        idx = np.append(idx, string)
    return idx


def col_Best(n):
    idx = np.array([])
    for i in np.arange(n):
        string = "B(:," + str(i + 1) + ')'
        idx = np.append(idx, string)
    return idx


def print_out(n,T,out_SVAR):
    estimator = out_SVAR['options']['estimator']
    loss = np.round(out_SVAR['loss'], 2)

    try:
        WUpdateSteps = out_SVAR['options']['kstep']
        if estimator == 'GMMcont':
            WUpdateSteps = 'CUE'
    except:
        WUpdateSteps = '-'


    try:
        scale = str()
        for i in range(n):
            scale = scale + str(np.round( np.diag(out_SVAR['options']['scale'])[i], 2))
            if i != n - 1:
                scale = scale + ' / '
    except:
        scale = '-'

    try:
        West = out_SVAR['options']['Wpara']
    except:
        West = '-'

    try:
        Avarest = out_SVAR['options']['Avarparametric']
    except:
        Avarest = '-'

    try:
        JTest = np.round(out_SVAR['J'], 2)
        JTestp = np.round(out_SVAR['Jpvalue'], 2)
    except:
        JTest = 'nan'
        JTestp = 'nan'
    try:
        Waldrec = np.round(out_SVAR['wald_rec'], 2)
        Waldrecp = np.round(out_SVAR['wald_rec_p'], 2)
    except:
        Waldrec = 'nan'
        Waldrecp = 'nan'
    try:
        moments_num = np.shape(out_SVAR['options']['moments'])[0]
        moments_num2 = np.sum(np.sum(out_SVAR['options']['moments'], axis=1) == 2)
        moments_num3 = np.sum(np.sum(out_SVAR['options']['moments'], axis=1) == 3)
        moments_num4 = np.sum(np.sum(out_SVAR['options']['moments'], axis=1) == 4)
    except:
        moments_num = 'nan'
        moments_num2 = 'nan'
        moments_num3 = 'nan'
        moments_num4 = 'nan'
    try:
        restrictions = out_SVAR['options']['restrictions']
        restrictions_num = np.power(n, 2) - np.sum(np.isnan(restrictions))
    except:
        restrictions = 'nan'
        restrictions_num = 'nan'

    try:
        omega = out_SVAR['omega']
        Ee2 = str()
        Ee3 = str()
        Ee4 = str()
        for i in range(n):
            Ee2 = Ee2 + str(np.round(omega[i, 1], 2))
            Ee3 = Ee3 + str(np.round(omega[i, 2], 2))
            Ee4 = Ee4 + str(np.round(omega[i, 3], 2))
            if i != n - 1:
                Ee2 = Ee2 + ' / '
                Ee3 = Ee3 + ' / '
                Ee4 = Ee4 + ' / '
    except:
        Ee2 = 'nan'
        Ee3 = 'nan'
        Ee4 = 'nan'

    try:
        moments_ident_num = np.shape(out_SVAR['options']['moments_ident'])[0]
        moments_ident_num2 = np.sum(np.sum(out_SVAR['options']['moments_ident'], axis=1) == 2)
        moments_ident_num3 = np.sum(np.sum(out_SVAR['options']['moments_ident'], axis=1) == 3)
        moments_ident_num4 = np.sum(np.sum(out_SVAR['options']['moments_ident'], axis=1) == 4)
    except:
        moments_ident_num = 'nan'
        moments_ident_num2 = 'nan'
        moments_ident_num3 = 'nan'
        moments_ident_num4 = 'nan'

    try:
        moments_overident_num = np.shape(out_SVAR['options']['moments_overident'])[0]
        moments_overident_num2 = np.sum(np.sum(out_SVAR['options']['moments_overident'], axis=1) == 2)
        moments_overident_num3 = np.sum(np.sum(out_SVAR['options']['moments_overident'], axis=1) == 3)
        moments_overident_num4 = np.sum(np.sum(out_SVAR['options']['moments_overident'], axis=1) == 4)
    except:
        moments_overident_num = 'nan'
        moments_overident_num2 = 'nan'
        moments_overident_num3 = 'nan'
        moments_overident_num4 = 'nan'

    try:
        moments_selected_num = np.shape(out_SVAR['options']['moments_overident'][out_SVAR['beta_est']==0])[0]
        moments_selected_num2 = np.sum(np.sum(out_SVAR['options']['moments_overident'][out_SVAR['beta_est']==0], axis=1) == 2)
        moments_selected_num3 = np.sum(np.sum(out_SVAR['options']['moments_overident'][out_SVAR['beta_est']==0], axis=1) == 3)
        moments_selected_num4 = np.sum(np.sum(out_SVAR['options']['moments_overident'][out_SVAR['beta_est']==0], axis=1) == 4)
    except:
        moments_selected_num = 'nan'
        moments_selected_num2 = 'nan'
        moments_selected_num3 = 'nan'
        moments_selected_num4 = 'nan'

    # TabData = list()
    # TabData.append([ estimator,  str(WUpdateSteps),   str(West) + '/' + str(Avarest) ,   str(BiasCorrection)  ])
    # tableSummary = tabulate(TabData, headers=np.array([
    #    str("Estimator:"   ), 'W update steps:', 'Estimator Wopt/Avar:',"Biascorrection:"]), tablefmt="github")
    # print(tableSummary)
    # print(" ")
    if estimator ==  'GMM':
        print('Estimator: GMM (steps='+ str(WUpdateSteps) + ')')
    elif estimator == 'CUE':
        print('Estimator: CUE')
    elif estimator == 'CSUE':
        print('Estimator: CSUE')
    elif estimator == 'GMM_W':
        print('Estimator: GMM white')
    elif estimator == 'GMM_WF':
        print('Estimator: GMM white fast')
    elif estimator == 'PML':
        print('Estimator: PML')
    elif estimator == 'LassoM':
        print('Estimator: LassoGMM (lambda=' +str(out_SVAR['options']['lambd']) + ')' )
    elif estimator == 'LassoMcont':
        print('Estimator: LassoCUE (lambda=' +str(out_SVAR['options']['lambd']) + ')' )
    elif estimator == 'LassoB':
        print('Estimator: LassoBGMM (lambda=' + str(out_SVAR['options']['lambd']) + ')')
    else:
        print("Unknown estimator")

    print('Estimator Wopt/Avar: ' + str(West) + '/' + str(Avarest))

    if estimator == 'LassoM' or estimator == 'LassoMcont':
        TabData = list()
        TabData.append(["#second: " + str(moments_ident_num2) + " /  " + str(moments_overident_num2) + " /  " + str(
            moments_selected_num2)])
        TabData.append(["#third: " + str(moments_ident_num3) + " /  " + str(moments_overident_num3) + " /  " + str(
            moments_selected_num3)])
        TabData.append(["#fourth: " + str(moments_ident_num4) + " /  " + str(moments_overident_num4) + " /  " + str(
            moments_selected_num4)])

        tableSummary = tabulate(TabData, headers=np.array([
            'Ident. / Overident. / Selected']), tablefmt="github")
        print(tableSummary)
        print(" ")

    # if estimator == 'GMM' or estimator == 'GMMcont' or estimator == 'GMM_W' or estimator == 'GMM_WF' or estimator == 'PML':
    TabData = list()
    TabData.append(["T=" + str(T), "#second: " + str(moments_num2), "WaldRec=" + str(Waldrec), "m=2: " + Ee2])
    TabData.append(["n=" + str(n), "#third: " + str(moments_num3), "WaldRec-pval=" + str(Waldrecp), "m=3: " + Ee3])
    TabData.append(["#restrictions:" + str(restrictions_num), "#fourth: " + str(moments_num4), "J=" + str(JTest),
                    "m=4: " + Ee4])
    TabData.append(
        ["#unknowns:" + str(np.power(n, 2) - restrictions_num), "->loss: " + str(loss), "J-pval=" + str(JTestp),
         " "  ])
    tableSummary = tabulate(TabData, headers=np.array([
       str("SVAR " ), 'Moments', 'Tests',"E[e_1^m] / ... / E[e_n^m]  "]), tablefmt="github")
    print(tableSummary)
    print(" ")


    # print("Moments of estimated structural shocks")
    # print_Moments(out_SVAR['omega'])
    # print(" ")

    TabData = list()
    b_est = np.round(SVAR.get_BVector(out_SVAR['B_est'], restrictions=restrictions, whiten=False), 2)

    try:
        Avar_est = np.diag(np.round(out_SVAR['Avar_est'], 2))
    except:
        Avar_est = np.full(np.shape(b_est), 'nan' )
    try:
        wald_all = np.round(out_SVAR['wald_all'], 2)
        wald_all_p = np.round(out_SVAR['wald_all_p'], 2)
    except:
        wald_all = np.full(np.shape(b_est), 'nan' )
        wald_all_p = np.full(np.shape(b_est), 'nan' )
    try:
        t_all = np.round(out_SVAR['t_all'], 2)
        t_all_p = np.round(out_SVAR['t_all_p'], 2)
    except:
        t_all = np.full(np.shape(b_est), 'nan' )
        t_all_p = np.full(np.shape(b_est), 'nan' )

    elementcounter = 0
    for i in range(n):
        for j in range(n):
            if not (np.isnan(restrictions[i, j])):
                b_est = np.insert(b_est, elementcounter, np.full(1, np.NaN), 0)
                wald_all = np.insert(wald_all, elementcounter, np.full(1, np.NaN), 0)
                wald_all_p = np.insert(wald_all_p, elementcounter, np.full(1, np.NaN), 0)
                t_all_p = np.insert(t_all_p, elementcounter, np.full(1, np.NaN), 0)
                t_all = np.insert(t_all, elementcounter, np.full(1, np.NaN), 0)
            elementcounter += 1
    rows = idx_Best(n)
    for i in range(n):
        this_row = list([rows[i]])
        for j in range(n):
            if np.isnan(b_est[n * i + j]):
                this_entry = '*' + str(np.round(restrictions[i, j],2)) + '*'
            else:
                this_entry = str(b_est[n * i + j])
            this_row.extend([this_entry])
        TabData.append(this_row)

        this_row = list([str("(avar/wald/pval)")])
        for j in range(n):
            this_entry = " (" + str(Avar_est[n * i + j]) + '/' + str(wald_all[n * i + j]) + '/' + str(
                wald_all_p[n * i + j]) + ')'
                         # '/' + str(t_all_p[n * i + j]) + \

            this_row.extend([this_entry])
        TabData.append(this_row)

        this_row = list()
        TabData.append(this_row)
    tableB = tabulate(TabData, headers=col_Best(n), stralign="center", tablefmt="github")
    print(tableB)
    print("Note: Restricted values are denoted by *x*")


def print_out_scaled(n,T,out_SVAR):
    estimator = out_SVAR['options']['estimator']+'_scaled'
    loss = np.round(out_SVAR['loss_scaled'], 2)
    try:
        JTest = np.round(out_SVAR['J_scaled'], 2)
        JTestp = np.round(out_SVAR['Jpvalue_scaled'], 2)
    except:
        JTest = 'nan'
        JTestp = 'nan'
    try:
        Waldrec = np.round(out_SVAR['wald_rec_scaled'], 2)
        Waldrecp = np.round(out_SVAR['wald_rec_p_scaled'], 2)
    except:
        Waldrec = 'nan'
        Waldrecp = 'nan'
    try:
        moments_num = np.shape(out_SVAR['options']['moments'])[0]
        moments_num2 = np.sum(np.sum(out_SVAR['options']['moments'], axis=1) == 2)
        moments_num3 = np.sum(np.sum(out_SVAR['options']['moments'], axis=1) == 3)
        moments_num4 = np.sum(np.sum(out_SVAR['options']['moments'], axis=1) == 4)
    except:
        moments_num = 'nan'
        moments_num2 = 'nan'
        moments_num3 = 'nan'
        moments_num4 = 'nan'
    try:
        restrictions = out_SVAR['options']['restrictions']
        restrictions_num = np.power(n, 2) - np.sum(np.isnan(restrictions))
    except:
        restrictions = 'nan'
        restrictions_num = 'nan'

    try:
        omega = out_SVAR['omega_scaled']
        Ee2 = str()
        Ee3 = str()
        Ee4 = str()
        for i in range(n):
            Ee2 = Ee2 + str(np.round(omega[i, 1], 2))
            Ee3 = Ee3 + str(np.round(omega[i, 2], 2))
            Ee4 = Ee4 + str(np.round(omega[i, 3], 2))
            if i != n - 1:
                Ee2 = Ee2 + ' / '
                Ee3 = Ee3 + ' / '
                Ee4 = Ee4 + ' / '
    except:
        Ee2 = 'nan'
        Ee3 = 'nan'
        Ee4 = 'nan'

    TabData = list()
    TabData.append(["T=" + str(T), "#second: " + str(moments_num2), "WaldRec=" + str(Waldrec) , "Notation: E[e_1^m] / ... / E[e_n^m]  "  ])
    TabData.append(["n=" + str(n), "#third: " + str(moments_num3), "WaldRec-pval=" + str(Waldrecp) , "m=2: " + Ee2])
    TabData.append(["#restrictions=" + str(restrictions_num), "#fourth: " + str(moments_num4), "J=" + str(JTest) , "m=3: " + Ee3])
    TabData.append(["#unknowns=" + str(np.power(n,2)-restrictions_num), "->loss: " + str(loss), "J-pval=" + str(JTestp), "m=4: " + Ee4])
    tableSummary = tabulate(TabData, headers=np.array([
       str("Estimator: " + estimator), '#moments', 'Tests','Moments of estimated shocks']), tablefmt="github")
    print(tableSummary)
    print(" ")

    # print("Moments of estimated structural shocks")
    # print_Moments(out_SVAR['omega'])
    # print(" ")

    TabData = list()
    b_est = np.round(SVAR.get_BVector(out_SVAR['B_est_scaled'], restrictions=restrictions, whiten=False), 2)

    try:
        Avar_est = np.diag(np.round(out_SVAR['Avar_est_scaled'], 2))
    except:
        Avar_est = np.full(np.shape(b_est), 'nan' )
    try:
        wald_all = np.round(out_SVAR['wald_all_scaled'], 2)
        wald_all_p = np.round(out_SVAR['wald_all_p_scaled'], 2)
    except:
        wald_all = np.full(np.shape(b_est), 'nan' )
        wald_all_p = np.full(np.shape(b_est), 'nan' )
    try:
        t_all = np.round(out_SVAR['t_all_scaled'], 2)
        t_all_p = np.round(out_SVAR['t_all_p_scaled'], 2)
    except:
        t_all = np.full(np.shape(b_est), 'nan' )
        t_all_p = np.full(np.shape(b_est), 'nan' )

    elementcounter = 0
    for i in range(n):
        for j in range(n):
            if not (np.isnan(restrictions[i, j])):
                b_est = np.insert(b_est, elementcounter, np.full(1, np.NaN), 0)
                wald_all = np.insert(wald_all, elementcounter, np.full(1, np.NaN), 0)
                wald_all_p = np.insert(wald_all_p, elementcounter, np.full(1, np.NaN), 0)
                t_all_p = np.insert(t_all_p, elementcounter, np.full(1, np.NaN), 0)
                t_all = np.insert(t_all, elementcounter, np.full(1, np.NaN), 0)
            elementcounter += 1
    rows = idx_Best(n)
    for i in range(n):
        this_row = list([rows[i]])
        for j in range(n):
            if np.isnan(b_est[n * i + j]):
                this_entry = '*' + str(restrictions[i, j]) + '*'
            else:
                this_entry = str(b_est[n * i + j])
            this_row.extend([this_entry])
        TabData.append(this_row)

        this_row = list([str("(avar/wald/pval)")])
        for j in range(n):
            this_entry = " (" + str(Avar_est[n * i + j]) + '/' + str(wald_all[n * i + j]) + '/' + str(
                wald_all_p[n * i + j]) + ')'
                         # '/' + str(t_all_p[n * i + j]) + \

            this_row.extend([this_entry])
        TabData.append(this_row)

        this_row = list()
        TabData.append(this_row)
    tableB = tabulate(TabData, headers=col_Best(n), stralign="center", tablefmt="github")
    print(tableB)
    print("Note: Restricted values are denoted by *x*")


def print_out_scaled(n,T,out_SVAR):
    estimator = out_SVAR['options']['estimator']+'_scaled'
    loss = np.round(out_SVAR['loss_scaled'], 2)
    try:
        JTest = np.round(out_SVAR['J_scaled'], 2)
        JTestp = np.round(out_SVAR['Jpvalue_scaled'], 2)
    except:
        JTest = 'nan'
        JTestp = 'nan'
    try:
        Waldrec = np.round(out_SVAR['wald_rec_scaled'], 2)
        Waldrecp = np.round(out_SVAR['wald_rec_p_scaled'], 2)
    except:
        Waldrec = 'nan'
        Waldrecp = 'nan'
    try:
        moments_num = np.shape(out_SVAR['options']['moments'])[0]
        moments_num2 = np.sum(np.sum(out_SVAR['options']['moments'], axis=1) == 2)
        moments_num3 = np.sum(np.sum(out_SVAR['options']['moments'], axis=1) == 3)
        moments_num4 = np.sum(np.sum(out_SVAR['options']['moments'], axis=1) == 4)
    except:
        moments_num = 'nan'
        moments_num2 = 'nan'
        moments_num3 = 'nan'
        moments_num4 = 'nan'
    try:
        restrictions = out_SVAR['options']['restrictions']
        restrictions_num = np.power(n, 2) - np.sum(np.isnan(restrictions))
    except:
        restrictions = 'nan'
        restrictions_num = 'nan'

    try:
        omega = out_SVAR['omega_scaled']
        Ee2 = str()
        Ee3 = str()
        Ee4 = str()
        for i in range(n):
            Ee2 = Ee2 + str(np.round(omega[i, 1], 2))
            Ee3 = Ee3 + str(np.round(omega[i, 2], 2))
            Ee4 = Ee4 + str(np.round(omega[i, 3], 2))
            if i != n - 1:
                Ee2 = Ee2 + ' / '
                Ee3 = Ee3 + ' / '
                Ee4 = Ee4 + ' / '
    except:
        Ee2 = 'nan'
        Ee3 = 'nan'
        Ee4 = 'nan'

    TabData = list()
    TabData.append(["T=" + str(T), "#second: " + str(moments_num2), "WaldRec=" + str(Waldrec) , "Notation: E[e_1^m] / ... / E[e_n^m]  "  ])
    TabData.append(["n=" + str(n), "#third: " + str(moments_num3), "WaldRec-pval=" + str(Waldrecp) , "m=2: " + Ee2])
    TabData.append(["#restrictions=" + str(restrictions_num), "#fourth: " + str(moments_num4), "J=" + str(JTest) , "m=3: " + Ee3])
    TabData.append(["#unknowns=" + str(np.power(n,2)-restrictions_num), "->loss: " + str(loss), "J-pval=" + str(JTestp), "m=4: " + Ee4])
    tableSummary = tabulate(TabData, headers=np.array([
       str("Estimator: " + estimator), '#moments', 'Tests','Moments of estimated shocks']), tablefmt="github")
    print(tableSummary)
    print(" ")

    # print("Moments of estimated structural shocks")
    # print_Moments(out_SVAR['omega'])
    # print(" ")

    TabData = list()
    b_est = np.round(SVAR.get_BVector(out_SVAR['B_est_scaled'], restrictions=restrictions, whiten=False), 2)

    try:
        Avar_est = np.diag(np.round(out_SVAR['Avar_est_scaled'], 2))
    except:
        Avar_est = np.full(np.shape(b_est), 'nan' )
    try:
        wald_all = np.round(out_SVAR['wald_all_scaled'], 2)
        wald_all_p = np.round(out_SVAR['wald_all_p_scaled'], 2)
    except:
        wald_all = np.full(np.shape(b_est), 'nan' )
        wald_all_p = np.full(np.shape(b_est), 'nan' )
    try:
        t_all = np.round(out_SVAR['t_all_scaled'], 2)
        t_all_p = np.round(out_SVAR['t_all_p_scaled'], 2)
    except:
        t_all = np.full(np.shape(b_est), 'nan' )
        t_all_p = np.full(np.shape(b_est), 'nan' )

    elementcounter = 0
    for i in range(n):
        for j in range(n):
            if not (np.isnan(restrictions[i, j])):
                b_est = np.insert(b_est, elementcounter, np.full(1, np.NaN), 0)
                wald_all = np.insert(wald_all, elementcounter, np.full(1, np.NaN), 0)
                wald_all_p = np.insert(wald_all_p, elementcounter, np.full(1, np.NaN), 0)
                t_all_p = np.insert(t_all_p, elementcounter, np.full(1, np.NaN), 0)
                t_all = np.insert(t_all, elementcounter, np.full(1, np.NaN), 0)
            elementcounter += 1
    rows = idx_Best(n)
    for i in range(n):
        this_row = list([rows[i]])
        for j in range(n):
            if np.isnan(b_est[n * i + j]):
                this_entry = '*' + str(restrictions[i, j]) + '*'
            else:
                this_entry = str(b_est[n * i + j])
            this_row.extend([this_entry])
        TabData.append(this_row)

        this_row = list([str("(avar/wald/pval)")])
        for j in range(n):
            this_entry = " (" + str(Avar_est[n * i + j]) + '/' + str(wald_all[n * i + j]) + '/' + str(
                wald_all_p[n * i + j]) + ')'
                         # '/' + str(t_all_p[n * i + j]) + \

            this_row.extend([this_entry])
        TabData.append(this_row)

        this_row = list()
        TabData.append(this_row)
    tableB = tabulate(TabData, headers=col_Best(n), stralign="center", tablefmt="github")
    print(tableB)
    print("Note: Restricted values are denoted by *x*")


def print_out_scaled(n,T,out_SVAR):
    estimator = out_SVAR['options']['estimator']+'_scaled'
    loss = np.round(out_SVAR['loss_scaled'], 2)
    try:
        JTest = np.round(out_SVAR['J_scaled'], 2)
        JTestp = np.round(out_SVAR['Jpvalue_scaled'], 2)
    except:
        JTest = 'nan'
        JTestp = 'nan'
    try:
        Waldrec = np.round(out_SVAR['wald_rec_scaled'], 2)
        Waldrecp = np.round(out_SVAR['wald_rec_p_scaled'], 2)
    except:
        Waldrec = 'nan'
        Waldrecp = 'nan'
    try:
        moments_num = np.shape(out_SVAR['options']['moments'])[0]
        moments_num2 = np.sum(np.sum(out_SVAR['options']['moments'], axis=1) == 2)
        moments_num3 = np.sum(np.sum(out_SVAR['options']['moments'], axis=1) == 3)
        moments_num4 = np.sum(np.sum(out_SVAR['options']['moments'], axis=1) == 4)
    except:
        moments_num = 'nan'
        moments_num2 = 'nan'
        moments_num3 = 'nan'
        moments_num4 = 'nan'
    try:
        restrictions = out_SVAR['options']['restrictions']
        restrictions_num = np.power(n, 2) - np.sum(np.isnan(restrictions))
    except:
        restrictions = 'nan'
        restrictions_num = 'nan'

    try:
        omega = out_SVAR['omega_scaled']
        Ee2 = str()
        Ee3 = str()
        Ee4 = str()
        for i in range(n):
            Ee2 = Ee2 + str(np.round(omega[i, 1], 2))
            Ee3 = Ee3 + str(np.round(omega[i, 2], 2))
            Ee4 = Ee4 + str(np.round(omega[i, 3], 2))
            if i != n - 1:
                Ee2 = Ee2 + ' / '
                Ee3 = Ee3 + ' / '
                Ee4 = Ee4 + ' / '
    except:
        Ee2 = 'nan'
        Ee3 = 'nan'
        Ee4 = 'nan'

    TabData = list()
    TabData.append(["T=" + str(T), "#second: " + str(moments_num2), "WaldRec=" + str(Waldrec) , "Notation: E[e_1^m] / ... / E[e_n^m]  "  ])
    TabData.append(["n=" + str(n), "#third: " + str(moments_num3), "WaldRec-pval=" + str(Waldrecp) , "m=2: " + Ee2])
    TabData.append(["#restrictions=" + str(restrictions_num), "#fourth: " + str(moments_num4), "J=" + str(JTest) , "m=3: " + Ee3])
    TabData.append(["#unknowns=" + str(np.power(n,2)-restrictions_num), "->loss: " + str(loss), "J-pval=" + str(JTestp), "m=4: " + Ee4])
    tableSummary = tabulate(TabData, headers=np.array([
       str("Estimator: " + estimator), '#moments', 'Tests','Moments of estimated shocks']), tablefmt="github")
    print(tableSummary)
    print(" ")

    # print("Moments of estimated structural shocks")
    # print_Moments(out_SVAR['omega'])
    # print(" ")

    TabData = list()
    b_est = np.round(SVAR.get_BVector(out_SVAR['B_est_scaled'], restrictions=restrictions, whiten=False), 2)

    try:
        Avar_est = np.diag(np.round(out_SVAR['Avar_est_scaled'], 2))
    except:
        Avar_est = np.full(np.shape(b_est), 'nan' )
    try:
        wald_all = np.round(out_SVAR['wald_all_scaled'], 2)
        wald_all_p = np.round(out_SVAR['wald_all_p_scaled'], 2)
    except:
        wald_all = np.full(np.shape(b_est), 'nan' )
        wald_all_p = np.full(np.shape(b_est), 'nan' )
    try:
        t_all = np.round(out_SVAR['t_all_scaled'], 2)
        t_all_p = np.round(out_SVAR['t_all_p_scaled'], 2)
    except:
        t_all = np.full(np.shape(b_est), 'nan' )
        t_all_p = np.full(np.shape(b_est), 'nan' )

    elementcounter = 0
    for i in range(n):
        for j in range(n):
            if not (np.isnan(restrictions[i, j])):
                b_est = np.insert(b_est, elementcounter, np.full(1, np.NaN), 0)
                wald_all = np.insert(wald_all, elementcounter, np.full(1, np.NaN), 0)
                wald_all_p = np.insert(wald_all_p, elementcounter, np.full(1, np.NaN), 0)
                t_all_p = np.insert(t_all_p, elementcounter, np.full(1, np.NaN), 0)
                t_all = np.insert(t_all, elementcounter, np.full(1, np.NaN), 0)
            elementcounter += 1
    rows = idx_Best(n)
    for i in range(n):
        this_row = list([rows[i]])
        for j in range(n):
            if np.isnan(b_est[n * i + j]):
                this_entry = '*' + str(restrictions[i, j]) + '*'
            else:
                this_entry = str(b_est[n * i + j])
            this_row.extend([this_entry])
        TabData.append(this_row)

        this_row = list([str("(avar/wald/pval)")])
        for j in range(n):
            this_entry = " (" + str(Avar_est[n * i + j]) + '/' + str(wald_all[n * i + j]) + '/' + str(
                wald_all_p[n * i + j]) + ')'
                         # '/' + str(t_all_p[n * i + j]) + \

            this_row.extend([this_entry])
        TabData.append(this_row)

        this_row = list()
        TabData.append(this_row)
    tableB = tabulate(TabData, headers=col_Best(n), stralign="center", tablefmt="github")
    print(tableB)
    print("Note: Restricted values are denoted by *x*")